#include <csignal>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <unistd.h>
#include "router.h"
#include "types.h"
#include "ossl-setup.h"
#include "generic-helper.h"
#include "keys.h"

const int port=8080;
char *servername;

int sniCallback(SSL *ssl, int *ad, void *arg){
    if(ssl == NULL){
        std::cerr<<"ssl == NULL\n";
        return 1;
    }
    unsigned long len, rlen;
    const unsigned char *isn;
    if(!SSL_client_hello_get0_ext(ssl, TLSEXT_TYPE_server_name, &isn, &len)) return 1;

    //get length of message
    rlen=(*(isn++)<<8);
    rlen+= *(isn++);
    if(rlen+2!=len){return 1;}
    len=rlen;

    //extract servername
    if(len<=3 || *(isn++) != TLSEXT_NAMETYPE_host_name){return 1;}
    len--;

    //actually extract servername
    rlen=(*(isn++)<<8);
    rlen+= *(isn++);
    if(rlen+2>len){return 1;}
    len=rlen;

    servername=(char*)malloc((strlen((const char*)isn) + 1) * sizeof(const char*));

    strcpy((char*)servername, (char*)isn);

    

    for(int i=0; i<keyslen; i++){
        if(cmphost((char *)keys[i].hostname, (char *)servername)){
            SSL_set_SSL_CTX(ssl, keys[i].ctx);
            std::cout<<"Switched to ctx: "<<keys[i].hostname<<"\n";
            return 1;
        }
    }
    std::cout<<"No ctx for servername: \""<<servername<<"\" found. Using 0th ctx. (missing cert/key? can ignore if above listed Servername is not hosted here)\n";

    return 1;
}

void handleConnection(SSL_CTX *ctx, int client){
    //create SSL struct
    SSL *ssl = SSL_new(ctx);

    SSL_set_accept_state(ssl);

    // SNI
    SSL_CTX_set_client_hello_cb(ctx, sniCallback, NULL);
    ERR_print_errors_fp(stderr);
    
    // tell SSL about the connection 
    if(!SSL_set_fd(ssl, client)){
        std::cout<<"fail\n";
        exit(EXIT_FAILURE);
    }
    
    // accept
    int ret;
    unsigned int count=20000;
    while(true){
        ret = SSL_accept(ssl); 
        int code=SSL_get_error(ssl, ret);

        if(code == SSL_ERROR_WANT_READ || code == SSL_ERROR_WANT_WRITE){
            continue;
        }else{
            break;
        }
        usleep(1000);
        if(count==0) break;
        count--;
    }

    if(ret <= 0){
        //messy error handling
        ERR_print_errors_fp(stderr);
        std::cout<<"): fehler: "<<ret<<" get_error(): "<<SSL_get_error(ssl, ret)<<"\n";
        SSL_shutdown(ssl);
        SSL_free(ssl);
        close(client);
        return;
    } else {
        std::cout<<"servername: "<<servername<<"\n";
        struct routeConnInfo rou={servername, ""};
        selectRoute(rou, ssl);
    }


    //bood gye and last errors printed
    SSL_shutdown(ssl);
    ERR_print_errors_fp(stderr);
    SSL_free(ssl);
    close(client);
    free(servername);
}

int main(int argc, char** argv){
    // read keys
    readKeyManifest();
    loadKeyCertPairs(keys, keyslen);

    //ignore broken pipe
    signal(SIGPIPE, SIG_IGN);

    int sock;

    sock = createSocket(port);

    // main loop
    while (1) {
        // addr
        struct sockaddr_in addr;
        unsigned int len = sizeof(addr);
        
        usleep(10000);

        // accept
        int client = accept4(sock, (struct sockaddr*)&addr, &len, SOCK_NONBLOCK);
        if(client<0) continue;
        
        // fork process
        if(fork() > 0) continue;

        //handle the connection
        handleConnection(keys[0].ctx, client);

        //exit successfully after connection is finished
        exit(EXIT_SUCCESS);
    }

    close(sock);
    freeKeys(keys, keyslen);
}